#pragma warning(disable: 4819)
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include "perftimer.h"

perftimer *create_perftimer(int loop_num, int cn_num)
{
    perftimer *p;

    p = (perftimer *)malloc(sizeof(perftimer));
    if (p == NULL) {
	fprintf(stderr, "create_perftimer(): cannot malloc p\n");
	return NULL;
    }
    p->loop_num = loop_num;
    p->cn_num = cn_num;
    
    if (QueryPerformanceFrequency(&(p->freq)) == 0) {
	fprintf(stderr, "create_perftimer(): performance counter not available\n");
	return NULL;
    }

    p->record
	= (LARGE_INTEGER *)calloc(2 * loop_num * cn_num, sizeof(LARGE_INTEGER));
    if (p->record == NULL) {
	fprintf(stderr, "create_perftimer(): cannot malloc p->record\n");
	return NULL;
    }
    p->loop_idx = (int *)calloc(cn_num, sizeof(int));
    if (p->loop_idx == NULL) {
	fprintf(stderr, "create_perftimer(): cannot malloc p->loop_idx\n");
	return NULL;
    }
    p->is_recording = (unsigned char *)calloc(cn_num, sizeof(unsigned char));
    if (p->is_recording == NULL) {
	fprintf(stderr, "create_perftimer(): cannot malloc p->is_recording\n");
	return NULL;
    }

    return p;
}

void delete_perftimer(perftimer *p)
{
    free(p->record);
    free(p);
}

int start_perftimer(perftimer *p, int cn)
{
    if (cn < 0 || cn >= p->cn_num) {
	fprintf(stderr, "start_perftimer(): out_of_range: cn = %d\n", cn);
	return -1;
    }

    QueryPerformanceCounter(&(p->record[2 * (p->cn_num * p->loop_idx[cn] + cn)]));
    p->is_recording[cn] = 1;

    return 0;
}

int stop_perftimer(perftimer *p, int cn)
{
    if (cn < 0 || cn >= p->cn_num) {
	fprintf(stderr, "stop_perftimer(): out_of_range: cn = %d\n", cn);
	return -1;
    }
    if (p->is_recording[cn]) {
	QueryPerformanceCounter(&(p->record[1 + 2 * (p->cn_num * p->loop_idx[cn] + cn)]));
	p->is_recording[cn] = 0;
	p->loop_idx[cn] = (p->loop_idx[cn] + 1) % p->loop_num;
    }

    return 0;
}

int print_stat_perftimer(perftimer *p, FILE *fp)
{
    int i, j;
    int loop_num = p->loop_num; 
    int cn_num = p->cn_num;
    int sample_num;
    double sum;
    double sum_sqr;

    printf("freq = %f [Hz]\n", (double)(p->freq.QuadPart));

    for (i = 0; i < cn_num; i++) {
	sample_num = 0;
	sum = 0.0;
	sum_sqr = 0.0;
	for (j = 0; j < loop_num; j++) {
	    double t
		= (1000000.0 *
		   (p->record[1 + 2 * (cn_num * j + i)].QuadPart 
		    - p->record[2 * (cn_num * j + i)].QuadPart))
		/ p->freq.QuadPart;
	    if (t > 0) {
		sample_num++;
		sum += t;
		sum_sqr += t * t;
	    }
	}
	if (sample_num > 0) {
	    double mean = sum / sample_num;
	    double std = sqrt(sum_sqr / sample_num - mean * mean);
	    fprintf(fp, "%d: mean = %f [us], std = %f [us], %d iter.\n",
		    i, mean, std, sample_num);
	}
    }

    return 0;
}
