#pragma warning(disable: 4819)
//#pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#include <stdio.h>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"
#include "perftimer.h"

#define OPT_BASELINE
//#define OPT_NAIVE
//#define OPT_INTERCHANGE
//#define OPT_LUT
//#define OPT_MEMCPY
//#define OPT_INDEX
//#define OPT_FUSION
//#define OPT_UNROLL

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))
#define PIXVALC(iplimagep, x, y, c) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (3 * (x)) + c))

int b2gray[256];
int g2gray[256];
int r2gray[256];

perftimer * pt;

int main()
{
    int i, j;
    int m, d;
    uchar r, g, b;
    int width, height;
    int wpad1, wpad3;
    int datasize1;
    uchar *cptr;
    uchar *mptr;
    uchar *bptr;
    int loop = 200;

    CvCapture *capture = cvCaptureFromCAM(0);

    IplImage *frame, *color, *mono, *back;
    pt = create_perftimer(loop, 1);

    for (i = 0; i < 256; i++) {
	b2gray[i] = 117 * i;
	g2gray[i] = 601 * i;
	r2gray[i] = 306 * i;
    }

    if (!capture) {
	fprintf(stderr, "cannot capture\n");
	return 1;
    }
    frame = cvQueryFrame(capture);
    color = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);  // color
    mono = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    back = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);

    width = color->width;
    height = color->height;
    wpad1 = mono->widthStep - width;
    wpad3 = color->widthStep - 3 * width;
    datasize1 = mono->widthStep * height;

    cvNamedWindow("image", CV_WINDOW_AUTOSIZE);

    while (1) {
	frame = cvQueryFrame(capture); /* cannot be modified */
	cvFlip(frame, color, 0);

	start_perftimer(pt, 0);

#if defined(OPT_BASELINE)
	cvCvtColor(color, mono, CV_BGR2GRAY);

	// frame diff
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}
	    
	cvCopy(mono, back, 0);
#elif defined(OPT_NAIVE)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		b = PIXVALC(color, i, j, 0);
		g = PIXVALC(color, i, j, 1);
		r = PIXVALC(color, i, j, 2);
		PIXVAL(mono, i, j) = (r * 306 + g * 601 + b * 117) / 1024;
	    }
	}

	// frame diff
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}
	    
	//cvCopy(mono, back, 0);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		PIXVAL(back, i, j) = PIXVAL(mono, i, j);
	    }
	}
#elif defined(OPT_INTERCHANGE)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		b = PIXVALC(color, i, j, 0);
		g = PIXVALC(color, i, j, 1);
		r = PIXVALC(color, i, j, 2);
		PIXVAL(mono, i, j) = (r * 306 + g * 601 + b * 117) / 1024;
	    }
	}

	// frame diff
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}
	    
	//cvCopy(mono, back, 0);
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		PIXVAL(back, i, j) = PIXVAL(mono, i, j);
	    }
	}
#elif defined(OPT_LUT)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		b = PIXVALC(color, i, j, 0);
		g = PIXVALC(color, i, j, 1);
		r = PIXVALC(color, i, j, 2);
		PIXVAL(mono, i, j) = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
	    }
	}

	// frame diff
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}
	    
	//cvCopy(mono, back, 0);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		PIXVAL(back, i, j) = PIXVAL(mono, i, j);
	    }
	}
#elif defined(OPT_MEMCPY)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		b = PIXVALC(color, i, j, 0);
		g = PIXVALC(color, i, j, 1);
		r = PIXVALC(color, i, j, 2);
		PIXVAL(mono, i, j) = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
	    }
	}

	// frame diff
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}

	//cvCopy(mono, back, 0);
	memcpy(back->imageData, mono->imageData, datasize1);
#elif defined(OPT_INDEX)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	cptr = (uchar *)(color->imageData);
	mptr = (uchar *)(mono->imageData); 
	bptr = (uchar *)(back->imageData);
	for (j = 0;
	     j < height;
	     j++, cptr += wpad3, mptr += wpad1, bptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i++, cptr += 3, mptr++, bptr++) {
		b = *cptr;
		g = *(cptr + 1);
		r = *(cptr + 2);
		*mptr = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
	    }
	}

	// frame diff
	cptr = (uchar *)(color->imageData);
	mptr = (uchar *)(mono->imageData);
	bptr = (uchar *)(back->imageData);
	for (j = 0;
	     j < height;
	     j++, cptr += wpad3, mptr += wpad1, bptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i++, cptr += 3, mptr++, bptr++) {
		m = *mptr;
		d = m - (*bptr);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *cptr = b;
		    *(cptr + 1) = g;
		    *(cptr + 2) = r;
		} else {
		    *cptr = m;
		    *(cptr + 1) = m;
		    *(cptr + 2) = m;
		}
	    }
	}

	//cvCopy(mono, back, 0);
	memcpy(back->imageData, mono->imageData, datasize1);
#elif defined(OPT_FUSION)
	cptr = (uchar *)(color->imageData);
	mptr = (uchar *)(mono->imageData);
	bptr = (uchar *)(back->imageData);
	for (j = 0;
	     j < height;
	     j++, cptr += wpad3, mptr += wpad1, bptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i++, cptr += 3, mptr++, bptr++) {
		//cvCvtColor(color, mono, CV_BGR2GRAY);
		b = *cptr;
		g = *(cptr + 1);
		r = *(cptr + 2);
		*mptr = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
		// frame diff
		m = *mptr;
		d = m - (*bptr);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *cptr = b;
		    *(cptr + 1) = g;
		    *(cptr + 2) = r;
		} else {
		    *cptr = m;
		    *(cptr + 1) = m;
		    *(cptr + 2) = m;
		}
	    }
	}

	//cvCopy(mono, back, 0);
	memcpy(back->imageData, mono->imageData, datasize1);
#elif defined(OPT_UNROLL)
	cptr = (uchar *)(color->imageData);
	mptr = (uchar *)(mono->imageData);
	bptr = (uchar *)(back->imageData);
	for (j = 0;
	     j < height;
	     j++, cptr += wpad3, mptr += wpad1, bptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i += 4, cptr += 12, mptr += 4, bptr += 4) {
		// [0]
		//cvCvtColor(color, mono, CV_BGR2GRAY);
		b = *cptr;
		g = *(cptr + 1);
		r = *(cptr + 2);
		m = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
		*mptr = m;
		// frame diff
		d = m - (*bptr);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *cptr = b;
		    *(cptr + 1) = g;
		    *(cptr + 2) = r;
		} else {
		    *cptr = m;
		    *(cptr + 1) = m;
		    *(cptr + 2) = m;
		}
		// [1]
		//cvCvtColor(color, mono, CV_BGR2GRAY);
		b = *(cptr + 3);
		g = *(cptr + 4);
		r = *(cptr + 5);
		m = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
		*(mptr + 1) = m;
		// frame diff
		d = m - (*(bptr + 1));
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *(cptr + 3) = b;
		    *(cptr + 4) = g;
		    *(cptr + 5) = r;
		} else {
		    *(cptr + 3) = m;
		    *(cptr + 4) = m;
		    *(cptr + 5) = m;
		}
		// [2]
		//cvCvtColor(color, mono, CV_BGR2GRAY);
		b = *(cptr + 6);
		g = *(cptr + 7);
		r = *(cptr + 8);
		m = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
		*(mptr + 2) = m;
		// frame diff
		d = m - (*(bptr + 2));
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *(cptr + 6) = b;
		    *(cptr + 7) = g;
		    *(cptr + 8) = r;
		} else {
		    *(cptr + 6) = m;
		    *(cptr + 7) = m;
		    *(cptr + 8) = m;
		}
		// [3]
		//cvCvtColor(color, mono, CV_BGR2GRAY);
		b = *(cptr + 9);
		g = *(cptr + 10);
		r = *(cptr + 11);
		m = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
		*(mptr + 3) = m;
		// frame diff
		d = m - (*(bptr + 3));
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *(cptr + 9) = b;
		    *(cptr + 10) = g;
		    *(cptr + 11) = r;
		} else {
		    *(cptr + 9) = m;
		    *(cptr + 10) = m;
		    *(cptr + 11) = m;
		}
	    }
	}

	//cvCopy(mono, back, 0);
	memcpy(back->imageData, mono->imageData, datasize1);
#endif

	stop_perftimer(pt, 0);

	cvShowImage("image", color);

	if (cvWaitKey(30) > 0) {
	    break;
	}
    }

    print_stat_perftimer(pt, stdout);

    return 0;
}
